/*
 *                            COPYRIGHT
 *
 *  PCB, interactive printed circuit board design
 *  Copyright (C) 1994,1995 Thomas Nau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses for paper mail and Email:
 *  Thomas Nau, Schlehenweg 15, 88471 Baustetten, Germany
 *  Thomas.Nau@rz.uni-ulm.de
 *
 */

static	char	*rcsid = "$Header: /sda7/users/nau/src/pcb/pcb-1.4.0/RCS/dev_ps.c,v 1.4 1996/01/13 16:35:33 nau Exp $";

/* PostScript device driver
 */
#include <math.h>
#include <time.h>

#include "global.h"

#include "data.h"
#include "dev_ps.h"
#include "error.h"
#include "misc.h"
#include "rotate.h"

/* ---------------------------------------------------------------------------
 * some local prototypes
 */
static	char	*PS_Init(PrintInitTypePtr Flags);
static	void	PS_Exit(void);
static	void	PS_SetColor(XColor);
static	void	PS_Drill(PinTypePtr);
static	void	PS_PrintLayer(LayerTypePtr, int);
static	void	PS_PrintLine(LineTypePtr);
static	void	PS_PrintPolygon(PolygonTypePtr);
static	void	PS_PrintText(TextTypePtr);
static	void	PS_PrintElementPackage(ElementTypePtr);
static	void	PS_PrintPad(PadTypePtr);
static	void	PS_PrintPinOrVia(PinTypePtr);

/* ----------------------------------------------------------------------
 * some local identifiers
 *
 * PSFunctions is initializes by the standard PostScript code used by PCB.
 * All PS functions are declared in this section:
 *   P:   draws a filled pin-polygon 
 *   PA:  draws a pad
 *   PV:  draws a pin or via
 *   PO:  draws a filled polygon 
 *   L:   draws a line
 *   B:   draws a filled box
 *   A:   draws an element arc
 * additional information is available from the PS comments.
 *
 * some PS code for arcs comes from the output of 'Fig2Dev'. Thanks to
 * the author
 */
static	char			*PS_Suffix[] = { "" };	/* suffix of filename */
static	PrintDeviceType	PS_QueryConstants = {
			"PostScript",
			PS_Init,
			PS_Exit,
			PS_SetColor,
			PS_Drill,
			PS_PrintLayer,
			PS_PrintElementPackage,
			PS_PrintPad,
			PS_PrintPinOrVia,
			PS_Suffix,
			1,					/* number of files */
			True,				/* driver can handle pipes */
			True};				/* driver can handle color */

static	PrintInitType	PS_Flags;

static	char	*PS_Functions[] = {
	"",
	"/PcbDict 200 dict def",
	"PcbDict begin",
	"PcbDict /DictMatrix matrix put",
	"",
	"% some constants",
	"/White {1.0 setgray} def",
	"/TAN {0.207106781} def",
	"/MTAN {-0.207106781} def",
	"",
	"% draw a filled polygon",
	"% get (x,y)... and number of points from stack",
	"/PO {",
	"	/number exch def",
	"	newpath",
	"	moveto",
	"	number 1 sub { lineto } repeat",
	"	closepath fill stroke",
	"} def",
	"",
	"/P {",
	"% draw a pin-polygon,",
	"% get x, y, length (lines), thickness and rotation from stack",
	"	/r exch def /thickness exch def /l exch def /y exch def /x exch def",
	"	gsave x y translate thickness thickness scale r rotate",
	"	newpath",
	"	0.5  MTAN",
	"	TAN  -0.5",
	"	MTAN -0.5",
	"	-0.5 MTAN",
	"	-0.5 TAN  l add",
	"	MTAN 0.5  l add",
	"	TAN  0.5  l add",
	"	0.5  TAN  l add",
	"	8 PO grestore",
	"} def",
	"",
	"/PV {",
	"% pin or via, x, y and thickness are on the stack",
	"	/drillinghole exch def /thickness exch def /y exch def /x exch def",
	"	x y 0 thickness 0 P",
	"% draw drilling whole	",
	"	White x y 0 drillinghole 0 P",
	"% restore drawing color",
	"   Color",
	"} def",
	"",
	"/L {",
	"% line, get x1, y1, x2, y2 and thickness from stack",
	"	/thick exch def /y2 exch def /x2 exch def /y1 exch def /x1 exch def",
	"	gsave thick setlinewidth",
	"	x1 y1 moveto x2 y2 lineto stroke",
	"	grestore",
	"} def",
	"",
	"/B {",
	"% filled box, get x1, y1, x2 and y2 from stack",
	"	/y2 exch def /x2 exch def /y1 exch def /x1 exch def",
	"	newpath",
	"	x1 y1 moveto x2 y1 lineto x2 y2 lineto x1 y2 lineto",
	"	closepath fill stroke",
	"} def",
	"",
	"/PA {",
	"% pad, same as line",
	"	L",
	"} def",
	"",
	"/A {",
	"% arc for elements, get x, y, width, height, thickness",
	"% startangle and delta-angle from stack",
	"/delta exch def /start exch def /thickness exch def",
	"/height exch def /width exch def /y exch def /x exch def",
	"% draw it",
	"	gsave thickness setlinewidth /save DictMatrix currentmatrix def",
	"% scaling is less then zero because the coord system has to be swapped",
	"	x y translate 0 width sub height scale",
	"	0 0 1 start start delta add arc save setmatrix stroke",
	"	grestore",
	"} def",
	"",
	"% setup some things",
	"0.0 setlinewidth",
	"1 setlinecap" };

/* ---------------------------------------------------------------------------
 * returns information about this driver
 */
PrintDeviceTypePtr PS_Query(void)
{
	return(&PS_QueryConstants);
}

/* ----------------------------------------------------------------------
 * prints PostScript header with function definition
 * info struct is are passed in
 */
static char *PS_Init(PrintInitTypePtr Flags)
{
	int			i;
	time_t		currenttime;
	Position	x1, y1,
				x2, y2;
	float		dx, dy;

		/* save passed-in data */
	PS_Flags = *Flags;
	currenttime = time(NULL);

		/* create standard PS header */
	fputs("%!PS-Adobe-2.0 EPSF-1.2\n", PS_Flags.FP[0]);
	fprintf(PS_Flags.FP[0], "%%%%Title: %s\n", UNKNOWN(PCB->Name));
	fprintf(PS_Flags.FP[0], "%%%%Creator: %s "RELEASE"\n", Progname);
	fprintf(PS_Flags.FP[0], "%%%%CreationDate: %s",
		asctime(localtime(&currenttime)));
	fputs("%%Orientation: Portrait\n", PS_Flags.FP[0]);

		/* rotate and scale bounding box to PS coordinates (1/72 inch) */
	x1 = (Position) ((float) PS_Flags.OffsetX * 0.072) -1;
	y1 = (Position) ((float) PS_Flags.OffsetY * 0.072) -1;
	dx = (float) (PS_Flags.BoundingBox.X2 -PS_Flags.BoundingBox.X1) *
		PS_Flags.Scale;
	dy = (float) (PS_Flags.BoundingBox.Y2 -PS_Flags.BoundingBox.Y1) *
		PS_Flags.Scale;;
	if (!PS_Flags.RotateFlag)
	{
		x2= (Position)((dx +PS_Flags.OffsetX) *0.072) +1;
		y2= (Position)((dy +PS_Flags.OffsetY) *0.072) +1;
	}
	else
	{
		x2= (Position)((dy +PS_Flags.OffsetX) *0.072) +1;
		y2= (Position)((dx +PS_Flags.OffsetY) *0.072) +1;
	}

		/* print BoundingBox header */
	fprintf(PS_Flags.FP[0], "%%%%BoundingBox: %i %i %i %i\n", x1, y1, x2, y2);
	fputs("%%Pages: 0\n", PS_Flags.FP[0]);
	fputs("%%BeginSetup\n", PS_Flags.FP[0]);
	fprintf(PS_Flags.FP[0], "%%%%IncludeFeature: *PageSize %s\n",
		PS_Flags.SelectedMedia->Name);
	fputs("%%EndSetup\n", PS_Flags.FP[0]);
	fputs("%%EndComments\n", PS_Flags.FP[0]);

		/* print own procedures */
	for (i = 0; i < ENTRIES(PS_Functions); i++)
		fprintf(PS_Flags.FP[0], "%s\n", PS_Functions[i]);

		/* add information about layout size, offset ... */
	fprintf(PS_Flags.FP[0], "%% PCBMIN(%d,%d), PCBMAX(%d,%d)\n",
		(int) PS_Flags.BoundingBox.X1, (int) PS_Flags.BoundingBox.Y1,
		(int) PS_Flags.BoundingBox.X2, (int) PS_Flags.BoundingBox.Y2);
	fprintf(PS_Flags.FP[0], "%% PCBOFFSET(%d,%d), PCBSCALE(%.5f)\n",
		PS_Flags.OffsetX, PS_Flags.OffsetY, PS_Flags.Scale);

		/* print first label */
	fputs("% PCBSTARTDATA --- do not remove ---\n", PS_Flags.FP[0]);
	fputs("gsave\n", PS_Flags.FP[0]);

		/* now insert transformation commands (reverse order):
		 * - move upper/left edge of layout to (0,0)
		 * - mirror it to transform X to PostScript coordinates
		 * - move to (0,0) again
		 * - if rotation is required, rotate and move to (0,0)
		 * - apply user scaling
		 * - move to new offset
		 * - scale to PostScript (72 dots per inch)
		 */ 
	fputs("0.072 0.072 scale\n", PS_Flags.FP[0]);
	fprintf(PS_Flags.FP[0], "%i %i translate\n", PS_Flags.OffsetX,
		PS_Flags.OffsetY);
	fprintf(PS_Flags.FP[0], "%.3f %.3f scale\n", PS_Flags.Scale,
		PS_Flags.Scale);
	if (PS_Flags.RotateFlag)
	{
		fprintf(PS_Flags.FP[0], "%i 0 translate\n",
			PS_Flags.BoundingBox.Y2 -PS_Flags.BoundingBox.Y1);
		fputs("90 rotate\n", PS_Flags.FP[0]);
	}
	if (PS_Flags.MirrorFlag)
	{
		fprintf(PS_Flags.FP[0], "0 %i translate\n",
			PS_Flags.BoundingBox.Y2 -PS_Flags.BoundingBox.Y1);
		fputs("1 -1 scale\n", PS_Flags.FP[0]);
	}
	fprintf(PS_Flags.FP[0], "%i %i translate\n", -PS_Flags.BoundingBox.X1,
		-PS_Flags.BoundingBox.Y1);

		/* black is default drawing color */
	fputs("/Color {0.0 setgray} def Color\n", PS_Flags.FP[0]);
	return(NULL);
}

/* ---------------------------------------------------------------------------
 * exit code for this driver is empty
 */
static void PS_Exit(void)
{
		/* print trailing commands */
	fputs("grestore\n", PS_Flags.FP[0]);
	fputs("% PCBENDDATA --- do not remove ---\n", PS_Flags.FP[0]);
/* done like this - rls */
#if 0
	fputs("showpage\n\004", PS_Flags.FP[0]);
#else
	fputs("showpage\n", PS_Flags.FP[0]);
#endif
}

/* ----------------------------------------------------------------------
 * handles drilling information
 * !!! not needed for this driver !!!
 */
static void PS_Drill(PinTypePtr PinOrVia)
{
		/* just get rid of warnings */
	PinOrVia = PinOrVia;
}

/* ----------------------------------------------------------------------
 * prints layer data
 */
static void PS_PrintLayer(LayerTypePtr Layer, int GroupNumber)
{
	fprintf(PS_Flags.FP[0], "%% LayerGroup: %i, Layer: \"%s\" (#%i)\n",
		GroupNumber, UNKNOWN(Layer->Name), GetLayerNumber(PCB->Data, Layer));
	LINE_LOOP(Layer, PS_PrintLine(line));
	TEXT_LOOP(Layer, PS_PrintText(text));
	POLYGON_LOOP(Layer, PS_PrintPolygon(polygon));
}

/* ----------------------------------------------------------------------
 * prints a line
 */
static void PS_PrintLine(LineTypePtr Line)
{
	fprintf(PS_Flags.FP[0], "%d %d %d %d %d L\n",
		(int) Line->Point1.X, (int) Line->Point1.Y,
		(int) Line->Point2.X, (int) Line->Point2.Y,
		(int) Line->Thickness);
}

/* ---------------------------------------------------------------------------
 * prints a filled polygon
 */
static void PS_PrintPolygon(PolygonTypePtr Ptr)
{
	int i = 0;

	POLYGONPOINT_LOOP(Ptr,
		{
			if (i++ % 9 == 8)
				fputc('\n', PS_Flags.FP[0]);
			fprintf(PS_Flags.FP[0], "%i %i ",
				(int) point->X, (int) point->Y);
		}
	);
	fprintf(PS_Flags.FP[0], "%d PO\n", Ptr->PointN);
}

/* ----------------------------------------------------------------------
 * prints a text
 * the routine is identical to DrawText() in module draw.c except
 * that DrawLine() and DrawRectangle() are replaced by their corresponding
 * printing routines
 */
static void PS_PrintText(TextTypePtr Text)
{
	Position		x = 0,
					width;
	unsigned char	*string = (unsigned char *) Text->TextString;
	Cardinal		n;
	FontTypePtr		font = &PCB->Font;

		/* get the center of the text for mirroring */
	width = Text->Direction & 0x01 ? 
				Text->BoundingBox.Y2 -Text->BoundingBox.Y1 :
				Text->BoundingBox.X2 -Text->BoundingBox.X1;
	while (string && *string)
	{
			/* draw lines if symbol is valid and data is present */
		if (*string <= MAX_FONTPOSITION && font->Symbol[*string].Valid)
		{
			LineTypePtr	line = font->Symbol[*string].Line;
			LineType	newline;

			for (n = font->Symbol[*string].LineN; n; n--, line++)
			{
					/* create one line, scale, move, rotate and swap it */
				newline = *line;
				newline.Point1.X = (newline.Point1.X +x) *Text->Scale /100;
				newline.Point1.Y = newline.Point1.Y      *Text->Scale /100;
				newline.Point2.X = (newline.Point2.X +x) *Text->Scale /100;
				newline.Point2.Y = newline.Point2.Y      *Text->Scale /100;
				newline.Thickness = newline.Thickness *Text->Scale /100;

					/* do some mirroring, swaping and rotating */
				if (TEST_FLAG(MIRRORFLAG, Text))
				{
					newline.Point1.X = width -newline.Point1.X;
					newline.Point2.X = width -newline.Point2.X;
				}
				RotateLineLowLevel(&newline, 0, 0, Text->Direction);

					/* the labels of SMD objects on the bottom
					 * side haven't been swapped yet, only their offset
					 */
				if (TEST_FLAG(ONBOTTOMFLAG, Text))
				{
					newline.Point1.X = SWAP_SIGN_X(newline.Point1.X);
					newline.Point1.Y = SWAP_SIGN_Y(newline.Point1.Y);
					newline.Point2.X = SWAP_SIGN_X(newline.Point2.X);
					newline.Point2.Y = SWAP_SIGN_Y(newline.Point2.Y);
				}
					/* add offset and draw line */
				newline.Point1.X += Text->X;
				newline.Point1.Y += Text->Y;
				newline.Point2.X += Text->X;
				newline.Point2.Y += Text->Y;
				PS_PrintLine(&newline);
			}

				/* move on to next cursor position */
			x += (font->Symbol[*string].Width +font->Symbol[*string].Delta);
		}
		else
		{
				/* the default symbol is a filled box */
			BoxType		defaultsymbol = PCB->Font.DefaultSymbol;
			Position	size = (defaultsymbol.X2 -defaultsymbol.X1) *6/5;

			defaultsymbol.X1 = (defaultsymbol.X1 +x) *Text->Scale /100;
			defaultsymbol.Y1 = defaultsymbol.Y1      *Text->Scale /100;
			defaultsymbol.X2 = (defaultsymbol.X2 +x) *Text->Scale /100;
			defaultsymbol.Y2 = defaultsymbol.Y2      *Text->Scale /100;

				/* do some mirroring ... */
			if (TEST_FLAG(MIRRORFLAG, Text))
			{
				defaultsymbol.X1 = width -defaultsymbol.X1;
				defaultsymbol.X2 = width -defaultsymbol.X2;
			}
			RotateBoxLowLevel(&defaultsymbol, 0, 0, Text->Direction);

				/* add offset and draw box */
			defaultsymbol.X1 += Text->X;
			defaultsymbol.Y1 += Text->Y;
			defaultsymbol.X2 += Text->X;
			defaultsymbol.Y2 += Text->Y;
			fprintf(PS_Flags.FP[0], "%d %d %d %d B\n",
				(int) defaultsymbol.X1, (int) defaultsymbol.Y1,
				(int) defaultsymbol.X2, (int) defaultsymbol.Y2);

				/* move on to next cursor position */
			x += size;
		}
		string++;
	}
}

/* ----------------------------------------------------------------------
 * prints package outline
 */
static void PS_PrintElementPackage(ElementTypePtr Element)
{
	ELEMENTLINE_LOOP(Element, PS_PrintLine(line););
	ARC_LOOP(Element, 
		fprintf(PS_Flags.FP[0], "%d %d %d %d %d %d %d A\n",
			(int) arc->X, (int) arc->Y,
			(int) arc->Width, (int) arc->Height,
			(int) arc->Thickness,
			arc->StartAngle,
			arc->Delta);
	);
	PS_PrintText(&ELEMENT_TEXT(PCB, Element));
}

/* ----------------------------------------------------------------------
 * prints a pad
 */
static void PS_PrintPad(PadTypePtr Pad)
{
	fprintf(PS_Flags.FP[0], "%d %d %d %d %d PA\n",
		(int) Pad->Point1.X, (int) Pad->Point1.Y,
		(int) Pad->Point2.X, (int) Pad->Point2.Y,
		(int) Pad->Thickness);
}

/* ----------------------------------------------------------------------
 * prints a via or pin
 */
static void PS_PrintPinOrVia(PinTypePtr Ptr)
{
	fprintf(PS_Flags.FP[0], "%d %d %d %d PV\n",
		(int) Ptr->X, (int) Ptr->Y,
		(int) Ptr->Thickness, (int) Ptr->DrillingHole);
}

/* ----------------------------------------------------------------------
 * queries color from X11 database and generates PostScript command
 * set PostScript identifier to the calculated value or to black by 
 * default
 */
static void PS_SetColor(XColor RGB)
{
	fprintf(PS_Flags.FP[0],
		"/Color {%.3f %.3f %.3f setrgbcolor} def Color\n",
		(float) RGB.red /65535.0,
		(float) RGB.green /65535.0,
		(float) RGB.blue /65535.0);
}
